@@ -1,6 +1,7 @@
 import torch
 import torch.nn.functional as F
 from torch import nn, einsum
+from torch.utils.checkpoint import checkpoint_sequential
 
 from en_transformer.rotary import SinusoidalEmbeddings, apply_rotary_pos_emb
 
@@ -131,7 +132,10 @@ class EquivariantAttention(nn.Module):
         norm_rel_coors = True,
         norm_coors_scale_init = 1.,
         use_cross_product = False,
-        talking_heads = False
+        talking_heads = False,
+        rotary_theta = 10000,
+        rel_dist_cutoff = 5000,
+        rel_dist_scale = 1e2
     ):
         super().__init__()
         self.scale = dim_head ** -0.5
@@ -189,8 +193,11 @@ class EquivariantAttention(nn.Module):
         num_coors_combine_heads = (2 if use_cross_product else 1) * heads
         self.coors_combine = nn.Parameter(torch.randn(num_coors_combine_heads))
 
-        self.rotary_emb = SinusoidalEmbeddings(dim_head // 2)
-        self.rotary_emb_seq = SinusoidalEmbeddings(dim_head // 2) if rel_pos_emb else None
+        self.rotary_emb = SinusoidalEmbeddings(dim_head // (2 if rel_pos_emb else 1), theta = rotary_theta)
+        self.rotary_emb_seq = SinusoidalEmbeddings(dim_head // 2, theta = rotary_theta) if rel_pos_emb else None
+
+        self.rel_dist_cutoff = rel_dist_cutoff
+        self.rel_dist_scale = rel_dist_scale
 
         self.init_eps = init_eps
         self.apply(self.init_)
@@ -239,7 +246,7 @@ class EquivariantAttention(nn.Module):
             nbhd_ranking = nbhd_ranking.masked_fill(self_mask, -1.)
             nbhd_ranking = nbhd_ranking.masked_fill(adj_mat, 0.)
 
-        if num_nn > 0:
+        if 0 < num_nn < n:
             # make sure padding does not end up becoming neighbors
             if exists(mask):
                 ranking_mask = mask[:, :, None] * mask[:, None, :]
@@ -289,7 +296,9 @@ class EquivariantAttention(nn.Module):
         rot_null = torch.zeros_like(rel_dist)
 
         q_pos_emb_rel_dist = self.rotary_emb(torch.zeros(n, device = device))
-        k_pos_emb_rel_dist = self.rotary_emb(rel_dist * 1e2)
+
+        rel_dist_to_rotate = (rel_dist * self.rel_dist_scale).clamp(max = self.rel_dist_cutoff)
+        k_pos_emb_rel_dist = self.rotary_emb(rel_dist_to_rotate)
 
         q_pos_emb = rearrange(q_pos_emb_rel_dist, 'i d -> () () i d')
         k_pos_emb = rearrange(k_pos_emb_rel_dist, 'b i j d -> b () i j d')
@@ -402,6 +411,18 @@ class EquivariantAttention(nn.Module):
 
 # transformer
 
+class Block(nn.Module):
+    def __init__(self, attn, ff):
+        super().__init__()
+        self.attn = attn
+        self.ff = ff
+
+    def forward(self, inp, coor_changes = None):
+        feats, coors, mask, edges, adj_mat = inp
+        feats, coors = self.attn(feats, coors, edges = edges, mask = mask, adj_mat = adj_mat)
+        feats, coors = self.ff(feats, coors)
+        return (feats, coors, mask, edges, adj_mat)
+
 class EnTransformer(nn.Module):
     def __init__(
         self,
@@ -424,7 +445,11 @@ class EnTransformer(nn.Module):
         norm_rel_coors = True,
         norm_coors_scale_init = 1.,
         use_cross_product = False,
-        talking_heads = False
+        talking_heads = False,
+        checkpoint = False,
+        rotary_theta = 10000,
+        rel_dist_cutoff = 5000,
+        rel_dist_scale = 1e2
     ):
         super().__init__()
         assert dim_head >= 32, 'your dimension per head should be greater than 32 for rotary embeddings to work well'
@@ -440,13 +465,14 @@ class EnTransformer(nn.Module):
         self.adj_emb = nn.Embedding(num_adj_degrees + 1, adj_dim) if exists(num_adj_degrees) and adj_dim > 0 else None
         adj_dim = adj_dim if exists(num_adj_degrees) else 0
 
+        self.checkpoint = checkpoint
         self.layers = nn.ModuleList([])
 
         for ind in range(depth):
-            self.layers.append(nn.ModuleList([
-                Residual(PreNorm(dim, EquivariantAttention(dim = dim, dim_head = dim_head, heads = heads, coors_hidden_dim = coors_hidden_dim, edge_dim = (edge_dim + adj_dim),  neighbors = neighbors, only_sparse_neighbors = only_sparse_neighbors, valid_neighbor_radius = valid_neighbor_radius, init_eps = init_eps, rel_pos_emb = rel_pos_emb, norm_rel_coors = norm_rel_coors, norm_coors_scale_init = norm_coors_scale_init, use_cross_product = use_cross_product, talking_heads = talking_heads))),
+            self.layers.append(Block(
+                Residual(PreNorm(dim, EquivariantAttention(dim = dim, dim_head = dim_head, heads = heads, coors_hidden_dim = coors_hidden_dim, edge_dim = (edge_dim + adj_dim),  neighbors = neighbors, only_sparse_neighbors = only_sparse_neighbors, valid_neighbor_radius = valid_neighbor_radius, init_eps = init_eps, rel_pos_emb = rel_pos_emb, norm_rel_coors = norm_rel_coors, norm_coors_scale_init = norm_coors_scale_init, use_cross_product = use_cross_product, talking_heads = talking_heads, rotary_theta = rotary_theta, rel_dist_cutoff = rel_dist_cutoff, rel_dist_scale = rel_dist_scale))),
                 Residual(PreNorm(dim, FeedForward(dim = dim)))
-            ]))
+            ))
 
     def forward(
         self,
@@ -489,15 +515,25 @@ class EnTransformer(nn.Module):
                 adj_emb = self.adj_emb(adj_indices)
                 edges = torch.cat((edges, adj_emb), dim = -1) if exists(edges) else adj_emb
 
+        assert not (return_coor_changes and self.training), 'you must be eval mode in order to return coordinates'
+
         # go through layers
 
         coor_changes = [coors]
+        inp = (feats, coors, mask, edges, adj_mat)
 
-        for attn, ff in self.layers:
-            feats, coors = attn(feats, coors, edges = edges, mask = mask, adj_mat = adj_mat)
-            coor_changes.append(coors)
+        # if in training mode and checkpointing is designated, use checkpointing across blocks to save memory
+        if self.training and self.checkpoint:
+            inp = checkpoint_sequential(self.layers, 1, inp)
+        else:
+            # iterate through blocks
+            for layer in self.layers:
+                inp = layer(inp)
+                coor_changes.append(inp[1]) # append coordinates for visualization
 
-            feats, coors = ff(feats, coors)
+        # return
+
+        feats, coors, *_ = inp
 
         if return_coor_changes:
             return feats, coors, coor_changes
