@@ -3,9 +3,9 @@ from torch import nn, einsum
 from einops import rearrange, repeat
 
 class SinusoidalEmbeddings(nn.Module):
-    def __init__(self, dim):
+    def __init__(self, dim, theta = 10000):
         super().__init__()
-        inv_freq = 1. / (10000 ** (torch.arange(0, dim, 2).float() / dim))
+        inv_freq = 1. / (theta ** (torch.arange(0, dim, 2).float() / dim))
         self.register_buffer('inv_freq', inv_freq)
 
     def forward(self, t):
